package router

import (
	"context"
	"errors"
	"time"

	"github.com/redis/rueidis"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/redistool"
	otelmetric "go.opentelemetry.io/otel/metric"
	"k8s.io/utils/clock"
)

const (
	tunnelsByAgentIDHashName = "tunnels_by_agent_id"
)

type Querier interface {
	// KASURLsByAgentID returns the list of kas URLs for a particular agent id.
	// A partial list may be returned together with an error.
	// Safe for concurrent use.
	KASURLsByAgentID(ctx context.Context, agentID int64) ([]string, error)
}

// RegistrationBuilder allows batching tunnel (un)registrations.
// Can be reused after Do is called.
type RegistrationBuilder interface {
	// Register registers tunnels for the given agent ids with the tracker.
	Register(ttl time.Duration, agentIDs ...int64)
	// Unregister unregisters tunnels for the given agent ids with the tracker.
	Unregister(agentIDs ...int64)
	// Refresh refreshes registered tunnels in the underlying storage.
	Refresh(ttl time.Duration, agentIDs ...int64)
	// Do executes the enqueued operations.
	Do(context.Context) error
}

// Registerer allows to register and unregister tunnels.
// Caller is responsible for periodically calling GC() and Refresh().
// Not safe for concurrent use.
type Registerer interface {
	RegistrationBuilder() RegistrationBuilder
	// GC deletes expired tunnels from the underlying storage.
	GC(ctx context.Context, agentIDs []int64) (int /* keysDeleted */, error)
}

type Tracker interface {
	Registerer
	Querier
}

type RedisTracker struct {
	ownPrivateAPIURL string
	clock            clock.PassiveClock
	tunnelsByAgentID redistool.ExpiringHashAPI[int64, string] // agentID -> kas URL -> nil
}

func NewRedisTracker(client rueidis.Client, agentKeyPrefix string, ownPrivateAPIURL string, m otelmetric.Meter) (*RedisTracker, error) {
	tunnelsByAgentID, err := redistool.NewRedisExpiringHashAPI[int64, string](tunnelsByAgentIDHashName, client, tunnelsByAgentIDHashKey(agentKeyPrefix), strToStr, m)
	if err != nil {
		return nil, err
	}
	return &RedisTracker{
		ownPrivateAPIURL: ownPrivateAPIURL,
		clock:            clock.RealClock{},
		tunnelsByAgentID: tunnelsByAgentID,
	}, nil
}

func (t *RedisTracker) RegistrationBuilder() RegistrationBuilder {
	return &redisRegistrationBuilder{
		ownPrivateAPIURL: t.ownPrivateAPIURL,
		clock:            t.clock,
		iob:              t.tunnelsByAgentID.IOBuilder(),
	}
}

func (t *RedisTracker) KASURLsByAgentID(ctx context.Context, agentID int64) ([]string, error) {
	var urls []string
	var errs []error
	err := t.tunnelsByAgentID.Scan(ctx, agentID, func(rawHashKey string, value []byte, err error) (bool, error) {
		if err != nil {
			errs = append(errs, err)
			return false, nil
		}
		urls = append(urls, rawHashKey)
		return false, nil
	})
	if err != nil {
		errs = append(errs, err)
	}
	return urls, errors.Join(errs...)
}

func (t *RedisTracker) GC(ctx context.Context, agentIDs []int64) (int /* keysDeleted */, error) {
	return t.tunnelsByAgentID.GCFor(agentIDs, true)(ctx)
}

type redisRegistrationBuilder struct {
	ownPrivateAPIURL string
	clock            clock.PassiveClock
	iob              redistool.IOBuilder[int64, string]
}

func (b *redisRegistrationBuilder) Register(ttl time.Duration, agentIDs ...int64) {
	b.iob.Set(agentIDs, ttl, kv(b.ownPrivateAPIURL, b.clock.Now().Add(ttl)))
}

func (b *redisRegistrationBuilder) Unregister(agentIDs ...int64) {
	b.iob.Unset(agentIDs, b.ownPrivateAPIURL)
}

func (b *redisRegistrationBuilder) Refresh(ttl time.Duration, agentIDs ...int64) {
	// Refresh is the same as registration right now, but this might change in the future.
	b.Register(ttl, agentIDs...)
}

func (b *redisRegistrationBuilder) Do(ctx context.Context) error {
	return b.iob.Do(ctx)
}

func kv(ownPrivateAPIURL string, expiresAt time.Time) redistool.BuilderKV[string] {
	return redistool.BuilderKV[string]{
		HashKey: ownPrivateAPIURL,
		Value: &redistool.ExpiringValue{
			ExpiresAt: expiresAt.Unix(),
			Value:     nil, // nothing to store.
		},
	}
}

// tunnelsByAgentIDHashKey returns a key for agentID -> (kasUrl -> nil).
func tunnelsByAgentIDHashKey(agentKeyPrefix string) redistool.KeyToRedisKey[int64] {
	prefix := agentKeyPrefix + ":kas_by_agent_id:"
	return func(agentID int64) string {
		return redistool.PrefixedInt64Key(prefix, agentID)
	}
}

func strToStr(key string) string {
	return key
}
