package ioz

import (
	"crypto/ed25519"
	"crypto/sha256"
	"encoding/base64"
	"fmt"
	"io"
	"log/slog"
	"os"
)

const (
	// hmacSHA256MinKeySize is the minimum HMAC SHA256 key size that is considered secure.
	// See https://www.rfc-editor.org/rfc/rfc2104#section-3 and https://en.wikipedia.org/wiki/HMAC.
	hmacSHA256MinKeySize = sha256.Size
	// hmacSHA256MaxKeySize is the maximum HMAC SHA256 key size that is not hashed into 32 bytes.
	hmacSHA256MaxKeySize = sha256.BlockSize
)

// DiscardData discards up to 8KiB of data from r.
// Function can be used to drain HTTP client response body.
// See https://pkg.go.dev/net/http#Response.
func DiscardData(r io.Reader) error {
	_, err := io.Copy(io.Discard, io.LimitReader(r, 8*1024))
	if err != nil {
		return fmt.Errorf("failed to read data: %w", err)
	}
	return nil
}

func LoadSHA256Base64Secret(log *slog.Logger, filename string) ([]byte, error) {
	return LoadBase64Secret(log, filename, hmacSHA256MinKeySize, hmacSHA256MaxKeySize)
}

func LoadEd25519Base64PublicKey(log *slog.Logger, filename string) ([]byte, error) {
	return LoadBase64Secret(log, filename, ed25519.PublicKeySize, ed25519.PublicKeySize)
}

func LoadBase64Secret(log *slog.Logger, filename string, minLen, maxLen int) ([]byte, error) {
	encodedAuthSecret, err := os.ReadFile(filename) //nolint: gosec
	if err != nil {
		return nil, fmt.Errorf("read file: %w", err)
	}
	decodedAuthSecret := make([]byte, base64.StdEncoding.DecodedLen(len(encodedAuthSecret)))

	n, err := base64.StdEncoding.Decode(decodedAuthSecret, encodedAuthSecret)
	if err != nil {
		return nil, fmt.Errorf("decoding: %w", err)
	}

	if n < minLen || n > maxLen {
		// TODO make this an error in 18.0, remove logging
		// TODO https://gitlab.com/gitlab-org/cluster-integration/gitlab-agent/-/issues/617
		log.Warn(fmt.Sprintf("secret in file %s is %d bytes long; allowed length is from %d to %d bytes", filename, n, minLen, maxLen))
		//return nil, fmt.Errorf("secret is %d bytes long; allowed length is from %d to %d bytes", n, minLen, maxLen)
	}

	return decodedAuthSecret[:n], nil
}
